import numpy as np
import xarray as xr
from matplotlib.colors import Colormap, LogNorm, Normalize

from ...utils import FileType
from ..color.colormap import Cmap, get_cmap


def get_default_norm(
    var: str,
    file_type: str | xr.Dataset | FileType | None = None,
) -> Normalize:

    if file_type is not None and not isinstance(file_type, FileType):
        file_type = FileType.from_input(file_type)

    if file_type == FileType.CPR_FMR_2A:
        if var in [
            "reflectivity_no_attenuation_correction",
            "reflectivity_corrected",
        ]:
            return Normalize(vmin=-40, vmax=20)
        elif var in [
            "brightness_temperature",
        ]:
            return Normalize(vmin=240, vmax=300)
        elif var in [
            "path_integrated_attenuation",
        ]:
            return Normalize(vmin=0, vmax=10)
    elif file_type == FileType.CPR_CD__2A:
        if var in [
            "doppler_velocity_uncorrected",
            "doppler_velocity_corrected_for_mispointing",
            "doppler_velocity_corrected_for_nubf",
            "doppler_velocity_integrated",
            "doppler_velocity_integrated_error",
            "doppler_velocity_best_estimate",
            "sedimentation_velocity_best_estimate",
            "sedimentation_velocity_best_estimate_error",
        ]:
            return Normalize(vmin=-6, vmax=6)
        elif var in [
            "spectrum_width_uncorrected",
            "spectrum_width_integrated",
            "spectrum_width_integrated_error",
        ]:
            return Normalize(vmin=0, vmax=5)
    elif file_type == FileType.CPR_CLD_2A:
        if var in ["water_content"]:
            return LogNorm(1e-6, 1e-3)
        elif var in ["characteristic_diameter"]:
            return LogNorm(1e-5, 2e-3)
        elif var in ["maximum_dimension_L"]:
            return LogNorm(1e-4, 2e-3)
        elif var in ["liquid_water_content"]:
            return LogNorm(1e-4, 1e-2)
        elif var in ["liquid_effective_radius"]:
            return Normalize(1e-4, 2e-3)
    elif file_type == FileType.ACM_CAP_2B:
        if var in ["ice_water_content"]:
            return LogNorm(1e-7, 1e-2)
        elif var in ["ice_effective_radius"]:
            return Normalize(0e-6, 200e-6)
        elif var in ["rain_water_content"]:
            return LogNorm(1e-3, 1e1)
        elif var in ["rain_median_volume_diameter"]:
            return Normalize(1e-5, 2e-3)
        elif var in ["liquid_water_content"]:
            return LogNorm(1e-7, 2e-3)
        elif var in ["liquid_effective_radius"]:
            return Normalize(0e-6, 50e-6)
        elif var in ["aerosol_extinction"]:
            return LogNorm(1e-7, 1e-3)
    elif file_type == FileType.MSI_CM__2A:
        if var in [
            "plot_cloud_mask_quality_status",
            "plot_cloud_type_quality_status",
            "plot_cloud_phase_quality_status",
        ]:
            return LogNorm(-0.5, 4.5)

    if var in [
        "mie_attenuated_backscatter",
        "rayleigh_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
    ]:
        return LogNorm(vmin=1e-8, vmax=1e-5)
    elif var in [
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
        "mie_total_attenuated_backscatter_355nm",
    ]:
        return LogNorm(vmin=1e-7, vmax=1e-4)
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        return LogNorm(vmin=1e-6, vmax=1e-3)
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        return Normalize(vmin=0, vmax=100)
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
    ]:
        return Normalize(vmin=0, vmax=0.6)
    elif var in [
        "plot_radarReflectivityFactor",
    ]:
        return Normalize(vmin=-25, vmax=20)
    elif var in [
        "plot_dopplerVelocity",
    ]:
        return Normalize(vmin=-2, vmax=4)
    elif "cloud_top_height_MSI" in var:
        return Normalize(vmin=0)
    elif var == "plot_cloud_top_height_difference_ATLID_MSI" and isinstance(
        file_type, xr.Dataset
    ):
        return Normalize(
            vmin=0,
            vmax=np.nanmax(file_type["cloud_top_height_difference_ATLID_MSI"].values),
        )
    elif "cloud_top_height_difference_ATLID_MSI" in var:
        return Normalize(vmin=0)
    elif "quality_status" in var:
        if file_type == FileType.MSI_CM__2A:
            return Normalize(vmin=-0.5, vmax=3.5)
        return Normalize(vmin=-1.5, vmax=4.5)
    elif var in [
        "ice_water_content",
    ]:
        return LogNorm(vmin=1e-4, vmax=5e-1)
    elif var in [
        "ice_effective_radius",
    ]:
        return Normalize(vmin=0, vmax=150)
    elif var in ["tir1", "tir2", "tir3"]:
        return Normalize(vmin=210, vmax=330)
    return Normalize()


def get_default_rolling_mean(
    var: str,
    file_type: str | xr.Dataset | FileType | None = None,
) -> int | None:

    if file_type is not None and not isinstance(file_type, FileType):
        file_type = FileType.from_input(file_type)

    if var in [
        "mie_attenuated_backscatter",
        "rayleigh_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
    ]:
        return 20
    if var in [
        "mie_total_attenuated_backscatter_355nm",
    ]:
        return 1
    return None


def get_default_cmap(
    var: str,
    file_type: str | xr.Dataset | FileType | None = None,
) -> Cmap:

    if file_type is not None and not isinstance(file_type, FileType):
        file_type = FileType.from_input(file_type)

    if file_type == FileType.CPR_CD__2A:
        if var in [
            "doppler_velocity_uncorrected",
            "doppler_velocity_corrected_for_mispointing",
            "doppler_velocity_corrected_for_nubf",
            "doppler_velocity_integrated",
            "doppler_velocity_integrated_error",
            "doppler_velocity_best_estimate",
            "sedimentation_velocity_best_estimate",
            "sedimentation_velocity_best_estimate_error",
        ]:
            return get_cmap("vik")
        elif var in [
            "spectrum_width_uncorrected",
            "spectrum_width_integrated",
            "spectrum_width_integrated_error",
        ]:
            return get_cmap("chiljet2")
    elif file_type == FileType.CPR_CLD_2A:
        if var in [
            "water_content",
            "characteristic_diameter",
            "maximum_dimension_L",
            "liquid_water_content",
            "liquid_effective_radius",
        ]:
            return get_cmap("chiljet2")
    elif file_type == FileType.AC__TC__2B:
        if var in [
            "synergetic_target_classification",
            "synergetic_target_classification_medium_resolution",
            "synergetic_target_classification_low_resolution",
        ]:
            return get_cmap("synergetic_tc")
    elif file_type == FileType.ACM_CAP_2B:
        if var in [
            "ice_water_content",
            "ice_effective_radius",
            "rain_water_content",
            "rain_median_volume_diameter",
            "liquid_water_content",
            "liquid_effective_radius",
            "aerosol_extinction",
        ]:
            return get_cmap("chiljet2")
    elif file_type == FileType.MSI_CM__2A:
        if var in [
            "plot_cloud_mask_quality_status",
            "plot_cloud_type_quality_status",
            "plot_cloud_phase_quality_status",
        ]:
            cmap = get_cmap("bam")
            colors = cmap(np.array([0.05, 0.3, 0.65, 0.9]))
            colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
            definitions = {v: str(v) for v in [0, 1, 2, 3, 4]}
            cmap = Cmap(colors, name="quality_status_amcth").to_categorical(definitions)
            return cmap
        elif var in ["cloud_mask"]:
            return get_cmap("msi_cloud_mask")
        elif var in ["cloud_phase"]:
            return get_cmap("msi_cloud_phase")
        elif var in ["plot_surface_classification"]:
            return get_cmap("msi_surface_classification")

    if var in [
        "mie_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
        "mie_total_attenuated_backscatter_355nm",
    ]:
        return get_cmap("calipso")
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        return get_cmap("chiljet2")
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        return get_cmap("chiljet2")
    elif var in [
        "rayleigh_attenuated_backscatter",
    ]:
        return get_cmap("ray")
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
    ]:
        return get_cmap("ratio")
    elif var in [
        "simple_classification",
    ]:
        return get_cmap("atl_simple_classification")
    elif var in [
        "classification",
        "classification_medium_resolution",
        "classification_low_resolution",
    ]:
        return get_cmap("atl_tc")
    elif var in [
        "plot_radarReflectivityFactor",
        "reflectivity_no_attenuation_correction",
        "reflectivity_corrected",
    ]:
        return get_cmap("radar_reflectivity")
    elif var in [
        "plot_dopplerVelocity",
    ]:
        return get_cmap("doppler_velocity")
    elif "cloud_top_height_MSI" in var:
        return get_cmap(get_cmap("navia").with_extremes(bad="#ffffff00"))
    elif "cloud_top_height_difference_ATLID_MSI" in var:
        return get_cmap(get_cmap("navia").with_extremes(bad="#808080", over="white"))
    elif "quality_status" in var:
        if isinstance(file_type, FileType):
            if file_type == FileType.AM__CTH_2B:
                cmap = get_cmap("roma_r")
                colors = cmap(np.linspace(0.1, 1, 5))
                colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
                definitions = {v: str(v) for v in [-1, 0, 1, 2, 3, 4]}
                cmap = Cmap(colors, name="quality_status_amcth").to_categorical(
                    definitions
                )
                return cmap
            elif file_type == FileType.CPR_TC__2A:
                cmap = get_cmap("roma_r")
                colors = cmap(np.linspace(0.1, 1, 5))
                colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
                definitions = {v: str(v) for v in [-1, 0, 1, 2, 3, 4]}
                cmap = Cmap(
                    ["#000000", "#BDBDBD"], name="quality_status_ctc"
                ).to_categorical({0: "good", 1: "bad"})
                return cmap
            elif file_type == FileType.MSI_CM__2A:
                cmap = get_cmap("roma_r")
                colors = cmap(np.linspace(0.1, 1, 4))
                # colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
                definitions = {v: str(v) for v in [0, 1, 2, 3]}
                cmap = Cmap(colors, name="quality_status_mcm").to_categorical(
                    definitions
                )
                return cmap
        cmap = get_cmap("roma_r")
        colors = cmap(np.linspace(0.1, 1, 5))
        colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
        definitions = {v: str(v) for v in [-1, 0, 1, 2, 3, 4]}
        cmap = Cmap(colors, name="quality_status_amcth").to_categorical(definitions)
        return cmap
    elif var in [
        "ice_water_content",
    ]:
        return get_cmap("chiljet2")
    elif var in [
        "ice_effective_radius",
    ]:
        return get_cmap("chiljet2")
    elif var in ["featuremask"]:
        return get_cmap("featuremask")
    elif file_type in [
        FileType.MSI_COP_2A,
        FileType.MSI_CM__2A,
    ] and var in [
        "cloud_type",
        "isccp_cloud_type",
    ]:
        return get_cmap("msi_cloud_type")
    elif var in ["hydrometeor_classification"]:
        return get_cmap("cpr_hydrometeor_classification")
    elif var in ["doppler_velocity_classification"]:
        return get_cmap("cpr_doppler_velocity_classification")
    elif var in ["simplified_convective_classification"]:
        return get_cmap("cpr_simplified_convective_classification")
    elif var in ["tir1", "tir2", "tir3"]:
        return get_cmap("Greys")
    return get_cmap("viridis")


def get_default_profile_range(
    var: str, ds: xr.Dataset | None = None
) -> tuple[float | None, float | None] | None:
    file_type: FileType | None = None
    if ds is not None and not isinstance(ds, FileType):
        file_type = FileType.from_input(ds)

    pad_frac = 0.00
    max_bsc = 8e-6  # [m-1 sr-1]
    max_ext = 6e-4  # [m-1]
    max_lr = 100.0  # [sr]
    max_depol = 0.5  # [-]

    _min: float
    _max: float
    if var in ["mie_attenuated_backscatter"]:
        _max = 0.8e-6
        return (-_max * pad_frac, _max)
    elif var in ["rayleigh_attenuated_backscatter"]:
        _max = 3e-6
        return (-_max * pad_frac, _max)
    elif var in ["crosspolar_attenuated_backscatter"]:
        _max = 0.2e-6
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
    ]:
        _max = max_bsc
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        _max = max_ext
        return (-_max * pad_frac, _max)
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        _max = max_lr
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
        "depol_ratio_from_means",
    ]:
        _max = max_depol
        return (-_max * pad_frac, _max)
    elif var in [
        "plot_radarReflectivityFactor",
    ]:
        _min = -25
        _max = 20
        vrange = _max - _min
        pad = vrange * pad_frac
        return (_min - pad, _max + pad)
    elif var in [
        "plot_dopplerVelocity",
    ]:
        _min = -2
        _max = 4
        vrange = _max - _min
        pad = vrange * pad_frac
        return (_min - pad, _max + pad)
    elif var in [
        "reflectivity_no_attenuation_correction",
        "reflectivity_corrected",
    ]:
        _min = -40
        _max = 20
        vrange = _max - _min
        pad = vrange * pad_frac
        return (_min - pad, _max + pad)
    return (None, None)



from typing import Final

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pyproj import Geod

_GEOD: Final[Geod] = Geod(ellps="WGS84")


def get_coord_between(
    coord1: ArrayLike,
    coord2: ArrayLike,
    f: float = 0.5,
) -> NDArray:
    """
    Interpolates between two coordinates by fraction f (0 to 1).

    Args:
        coord1 (ArrayLike): The first lat/lon point.
        coord2 (ArrayLike): The second lat/lon point.
        f (float): A fractional value between 0 and 1. Defaults to 0.5, i.e., the mid point between coord1 and coord2.

    Returns:
        NDArray: A 2-element `numpy.ndarray` representing the interpolated lat/lon point.
    """

    coord1 = np.array(coord1)
    coord2 = np.array(coord2)

    if coord1.shape != (2,):
        raise ValueError(f"coord1 must be a 2-element sequence (lat, lon)")

    if coord2.shape != (2,):
        raise ValueError(f"coord2 must be a 2-element sequence (lat, lon)")

    lon, lat = interpgeo(
        lat1=float(coord1[0]),
        lon1=float(coord1[1]),
        lat2=float(coord2[0]),
        lon2=float(coord2[1]),
        f=f,
    )
    return np.array([lat, lon])


def interpgeo(
    lat1: float,
    lon1: float,
    lat2: float,
    lon2: float,
    f: float,
) -> tuple[float, float]:
    """
    Interpolates along the geodesic from (lon1, lat1) to (lon2, lat2) by fraction f (0 to 1) and returns interpolated (lon, lat).
    """
    azi1, azi2, dist = _GEOD.inv(lon1, lat1, lon2, lat2)
    lon, lat, _ = _GEOD.fwd(lon1, lat1, azi1, f * dist)
    return lon, lat



from typing import Literal, TypeAlias

import matplotlib.ticker as ticker
import numpy as np
from matplotlib.axes import Axes

from .axis import AxisInput, validate_axis_input


def _smart_format(x, n=5):
    """Format with n decimals, then strip trailing 0 if present"""
    s = f"{x:.{n}f}"
    if s.endswith("0" * n):
        return s[: -n + 1]
    for i in range(n - 1, 0, -1):
        if s.endswith("0" * i):
            return s[:-i]
    return s


def format_height_ticks(
    ax: Axes,
    axis: AxisInput = "y",
    label: str | None = "Height",
    show_tick_labels: bool = True,
    show_units: bool = True,
):
    if not isinstance(ax, Axes):
        raise TypeError(
            f"{format_height_ticks.__name__}() for `ax` expected type '{Axes.__name__}' but got '{type(ax).__name__}' instead"
        )

    axis = validate_axis_input(axis)

    locator = ticker.MaxNLocator(nbins="auto", min_n_ticks=4, steps=[1, 2, 2.5, 5, 10])
    _ax = {"y": ax.yaxis, "x": ax.xaxis}[axis]
    _ax.set_major_locator(locator)

    height_ticks = ax.get_yticks() if axis == "y" else ax.get_xticks()
    height_lim = ax.get_ylim() if axis == "y" else ax.get_xlim()
    height_ticks = np.array(
        [h for h in height_ticks if h >= height_lim[0] and h <= height_lim[1]]
    )
    height_distance = height_ticks[-1] - height_ticks[0]
    units = ""
    _max_height_distance = 2.3e3

    if show_units and label != "none" and label is not None:
        units = " [km]" if height_distance >= _max_height_distance else " [m]"

    new_height_ticks = height_ticks
    if height_distance >= _max_height_distance:
        new_height_ticks = np.array([h / 1000 for h in height_ticks])

    all_ticks_are_integers = np.array([t.is_integer() for t in new_height_ticks]).all()
    if all_ticks_are_integers:
        new_height_ticks = np.array([int(t) for t in new_height_ticks])

    if not show_tick_labels:
        new_height_ticks = np.array(["" for t in new_height_ticks])

    if np.issubdtype(new_height_ticks.dtype, np.floating):
        new_height_ticks = np.array([_smart_format(t) for t in new_height_ticks])

    if axis == "y":
        ax.set_yticks(height_ticks, labels=new_height_ticks)
    else:
        ax.set_xticks(height_ticks, labels=new_height_ticks)

    if label == "none" or label is None:
        label = ""

    if axis == "y":
        ax.set_ylabel(f"{label}{units}")
    else:
        ax.set_xlabel(f"{label}{units}")



import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex


def get_cmap():
    name = "chiljet"
    colors = [
        "#DEDEDE",
        "#0000FF",
        "#009E9E",
        "#00FF00",
        "#FEFE00",
        "#FE7E00",
        "#FF0000",
        "#7E007E",
    ]
    positions = [
        0.0,
        0.234375,
        0.359375,
        0.484375,
        0.609375,
        0.734375,
        0.859375,
        1.0,
    ]
    color_dict = list(zip(positions, colors))
    cmap = LinearSegmentedColormap.from_list(name, color_dict)
    return cmap



from typing import Iterable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray
from xarray import Dataset

from ..constants import ALONG_TRACK_DIM, TIME_VAR, TRACK_LAT_VAR
from ..np_array_utils import flatten_array, pad_true_sequence
from ..time import TimeRangeLike, TimestampLike, to_timestamp
from ..typing import NumericPairLike, NumericPairNoneLike, validate_numeric_pair


def filter_index(
    ds: Dataset,
    index: int | slice | NDArray,
    along_track_dim: str = ALONG_TRACK_DIM,
    pad_idxs: int = 0,
) -> Dataset:
    """
    Filters a dataset given an along-track index number, list/array or range/slice.

    Args:
        ds (Dataset): Input dataset with along-track dimension.
        index (int | slice | NDArray): Index(es) to filter.
        along_track_dim (str, optional): Dimension along which to apply filtering. Defaults to ALONG_TRACK_DIM.
        pad_idxs (int, optional): Number of additional samples added at both sides of the selection.
            This input is ignored when `index` is an array-like. Defaults to 0.

    Returns:
        Dataset: Filtered dataset.
    """
    if isinstance(index, int):
        index = slice(index, index + 1)

    if isinstance(index, slice):
        index = slice(index.start - pad_idxs, index.stop + pad_idxs, index.step)
    else:
        index = flatten_array(index)

    return ds.isel({along_track_dim: index})



import re


def _parse_degrees_units(units: str, name: str, use_latex: bool = False) -> str | None:
    first_letter = name[0].upper()
    pattern = re.compile(
        rf"(?i)^(?:"
        rf"deg[_ ]?{first_letter}|"
        rf"degree[_ ]?{name}|"
        rf"{name}[_ ]?(?:deg|degrees)?|"
        rf"°\s*{first_letter}|"
        rf"\$\^\{{\\circ\}}\${first_letter}"
        rf")$",
        flags=re.UNICODE,
    )
    match = pattern.match(units)
    if match:
        if use_latex:
            return rf"$^{{\circ}}${first_letter}"
        return f"deg_{name.lower()}"
    return None


def parse_units(units: str, use_latex: bool = True) -> str:
    units = units.strip()

    pattern = re.compile(
        r"(?i)^((mm|msr)(\^?-1)\s*(m|sr)(\^?-1)|(m|sr)(\^?-1)\s*(mm|msr)(\^?-1))$"
    )
    match = pattern.match(units)
    if match:
        if use_latex:
            return "Mm$^{-1}$ sr$^{-1}$"
        return "Mm-1 sr-1"

    pattern = re.compile(r"(?i)^((m)(\^?-1)\s*(sr)(\^?-1)|(sr)(\^?-1)\s*(m)(\^?-1))$")
    match = pattern.match(units)
    if match:
        if use_latex:
            return "m$^{-1}$ sr$^{-1}$"
        return "m-1 sr-1"

    pattern = re.compile(r"(?i)^mm(\^?-1)$")
    match = pattern.match(units)
    if match:
        if use_latex:
            return "Mm$^{-1}$"
        return "Mm-1"

    pattern = re.compile(r"(?i)^m(\^?-1)$")
    match = pattern.match(units)
    if match:
        if use_latex:
            return "m$^{-1}$"
        return "m-1"

    pattern = re.compile(r"(?i)^msr(\^?-1)$")
    match = pattern.match(units)
    if match:
        if use_latex:
            return "Msr$^{-1}$"
        return "Msr-1"

    pattern = re.compile(r"(?i)^sr(\^?-1)$")
    match = pattern.match(units)
    if match:
        if use_latex:
            return "sr$^{-1}$"
        return "sr-1"

    pattern = re.compile(r"(?i)^dbz$")
    match = pattern.match(units)
    if match:
        return "dBZ"

    if s := _parse_degrees_units(units, "celsius", use_latex):
        return s

    if s := _parse_degrees_units(units, "fahrenheit", use_latex):
        return s

    if s := _parse_degrees_units(units, "north", use_latex):
        return s

    if s := _parse_degrees_units(units, "east", use_latex):
        return s

    if s := _parse_degrees_units(units, "south", use_latex):
        return s

    if s := _parse_degrees_units(units, "west", use_latex):
        return s

    if use_latex:
        units = re.sub(r"(?<!\$\^\{)\s*[-−]([0-9]+)", r"$^{-\1}$", units)
    else:
        units = re.sub(r"\$\^\{[-−]([0-9]+)\}\$", r"-\1", units)

    return units



from dataclasses import dataclass, field


@dataclass(frozen=True)
class GroundSite:
    """Class representing a geographic site (or ground station) with associated metadata.

    Attributes:
        latitude (float): Latitude of the site in decimal degrees.
        longitude (float): Longitude of the site in decimal degrees.
        name (str): Short name or identifier of the site.
        long_name (str): Full descriptive name of the site.
        aliases (list[str]): Alternative names or identifiers for the site.
        altitude (float): Altitude of the site in meters above sea level.
        cloudnet_name (str | None): Identifier string used in CloudNet file names, or None if not applicable.
    """

    latitude: float
    """Latitude of the site in decimal degrees."""
    longitude: float
    """Longitude of the site in decimal degrees."""
    name: str = ""
    """Short name or identifier of the site."""
    long_name: str = ""
    """Full descriptive name of the site."""
    aliases: list[str] = field(default_factory=list)
    """Alternative names or identifiers for the site."""
    altitude: float = 0.0
    """Altitude of the site in meters above sea level."""
    cloudnet_name: str | None = None
    """Identifier string used in CloudNet file names, or None if not applicable."""

    @property
    def coordinates(self) -> tuple[float, float]:
        """Geodetic coordinates of the ground site (lat,lon)."""
        return (self.latitude, self.longitude)


GROUND_SITES: list[GroundSite] = [
    GroundSite(
        name="TROPOS",
        long_name="TROPOS (LPZ)",
        aliases=["tropos"],
        latitude=51.352757,
        longitude=12.43392,
        altitude=125,
        cloudnet_name="leipzig",
    ),
    GroundSite(
        name="Leipzig",
        long_name="Leipzig (GER)",
        aliases=["leipzig", "lpz"],
        latitude=51.352757,
        longitude=12.43392,
        altitude=125,
        cloudnet_name="leipzig",
    ),
    GroundSite(
        name="Mindelo",
        long_name="Mindelo (CPV)",
        aliases=["mindelo", "cpv"],
        latitude=16.878,
        longitude=-24.995,
        altitude=13,
        cloudnet_name="mindelo",
    ),
    GroundSite(
        name="Dushanbe",
        long_name="Dushanbe (TJK)",
        aliases=["dushanbe", "tjk"],
        latitude=38.559,
        longitude=68.856,
        altitude=864,
        cloudnet_name="dushanbe",
    ),
    GroundSite(
        name="Melpitz",
        long_name="Melpitz (GER)",
        aliases=["melpitz"],
        latitude=51.526,
        longitude=12.928,
        altitude=83,
        cloudnet_name="melpitz",
    ),
    GroundSite(
        name="Limassol",
        long_name="Limassol (CYP)",
        aliases=["limassol", "cyp", "cyprus"],
        latitude=34.67667,
        longitude=33.04417,
        altitude=3,
        cloudnet_name="limassol",
    ),
    GroundSite(
        name="Antikythera",
        long_name="Antikythera (GRC)",
        aliases=["antikythera", "grc", "greece"],
        latitude=35.86,
        longitude=23.31,
        altitude=193,
        cloudnet_name="antikythera",
    ),
    GroundSite(
        name="Evora",
        long_name="Evora (PRT)",
        aliases=["evora", "prt", "portugal"],
        latitude=38.576,
        longitude=-7.911,
        altitude=290,
        cloudnet_name="evora",
    ),
    GroundSite(
        name="HPB",
        long_name="Hohenpeissenberg (GER)",
        aliases=["hpb", "hohenpeissenberg"],
        latitude=47.801473,
        longitude=11.009348,
        altitude=990,
        cloudnet_name="hpb",
    ),
    GroundSite(
        name="Warsaw",
        long_name="Warsaw (POL)",
        aliases=["warsaw", "pol", "poland"],
        latitude=52.233608,
        longitude=21.020265,
        altitude=113,
        cloudnet_name="warsaw",
    ),
    GroundSite(
        name="Kuopio",
        long_name="Kuopio (FIN)",
        aliases=["kuopio", "fin", "finland"],
        latitude=62.965698,
        longitude=27.666004,
        altitude=78,
        cloudnet_name="kuopio",
    ),
    GroundSite(
        name="Cabauw",
        long_name="Cabauw (NL)",
        aliases=["cabauw", "nl", "netherlands"],
        latitude=51.9677,
        longitude=4.9271,
        altitude=0,
        cloudnet_name=None,
    ),
    GroundSite(
        name="Koganei",
        long_name="Koganei (JP)",
        aliases=["koganei", "jp", "japan"],
        latitude=35.7,
        longitude=139.48,
        altitude=70,
        cloudnet_name=None,
    ),
    GroundSite(
        name="Neumayer III",
        long_name="Neumayer-Station III",
        aliases=["neumayer", "neumayer3", "neumayeriii"],
        latitude=-70.674444,
        longitude=-8.274167,
        altitude=41,
        cloudnet_name=None,
    ),
    GroundSite(
        name="Invercargill",
        long_name="Invercargill (NZ)",
        aliases=["invercargill", "gosouth"],
        latitude=-46.40000153,
        longitude=168.3000031,
        altitude=20,
        cloudnet_name=None,
    ),
]


def get_ground_site(site: str | GroundSite) -> GroundSite:
    """Retruns ground site data based on name and raises `ValueError` if no matching ground site is found and `TypeError`."""
    if isinstance(site, GroundSite):
        return site
    if not isinstance(site, str):
        raise TypeError(
            f"{get_ground_site.__name__}() Expected type `{str.__name__}` but got `{type(site).__name__}` (name={site})"
        )
    site = site.lower()
    for gs in GROUND_SITES:
        if site in gs.aliases:
            return gs

    gss = [gs.name for gs in GROUND_SITES]
    error_msg = f"""No matching ground site found: '{site}'. Supported site names are: '{gss[0]}', {"', '".join(gss[1:-1])}', and '{gss[-1]}'."""
    raise ValueError(error_msg)



from matplotlib.figure import Figure

from ...utils.typing import HasFigure


def save_figure_with_auto_margins(
    figure: Figure | HasFigure,
    filename: str,
    pad: float = 0.1,
    **kwargs,
) -> None:
    """
    Save a figure as an image or vector graphic to a file.

    Args:
        figure (Figure | HasFigure): A figure object (`matplotlib.figure.Figure`) or objects exposing a `.fig` attribute containing a figure (e.g., `CurtainFigure`).
        filename (str): The path where the image is saved.
        pad (float): Extra padding (i.e., empty space) around the image in inches. Defaults to 0.1.
        **kwargs (dict[str, Any]): Keyword arguments passed to wrapped function call of `matplotlib.pyplot.savefig`.
    """
    if isinstance(figure, Figure):
        fig = figure
    else:
        fig = figure.fig

    # Get original sizes
    original_size: tuple[float, float] = tuple(fig.get_size_inches())
    original_positions = {}
    for ax in fig.get_axes():
        original_positions[ax] = ax.get_position()

    fig.canvas.draw()

    # Get the bounding box of all figure elements in inches
    bbox = fig.get_tightbbox(fig.canvas.get_renderer())  # type: ignore

    # Calculate required padding on each side in inches
    pad_left = max(0, -bbox.x0)
    pad_right = max(0, bbox.x1 - original_size[0])
    pad_bottom = max(0, -bbox.y0)
    pad_top = max(0, bbox.y1 - original_size[1])

    # Add extra outside padding
    pad_left += pad
    pad_right += pad
    pad_bottom += pad
    pad_top += pad

    # Calculate new figure size
    new_width = original_size[0] + pad_left + pad_right
    new_height = original_size[1] + pad_bottom + pad_top

    # Resize figure
    fig.set_size_inches(new_width, new_height)

    # Reposition all axes to account for padding
    for ax in fig.get_axes():
        old_pos = original_positions[ax]
        ax.set_position(
            (
                (old_pos.x0 * original_size[0] + pad_left) / new_width,
                (old_pos.y0 * original_size[1] + pad_bottom) / new_height,
                (old_pos.width * original_size[0]) / new_width,
                (old_pos.height * original_size[1]) / new_height,
            )
        )

    # Save figure as an image
    fig.savefig(filename, **kwargs)

    # Restore original settings
    for ax in fig.get_axes():
        ax.set_position(original_positions[ax])
    fig.set_size_inches(original_size)



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_amacd(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens AM__ACD_2B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    nadir_idx = get_nadir_index(ds, nadir_idx=150)

    ds = ds.rename({"latitude": SWATH_LAT_VAR})
    ds = ds.rename({"longitude": SWATH_LON_VAR})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var=SWATH_LAT_VAR, swath_lon_var=SWATH_LON_VAR
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        time_var="time",
    )

    ds = ds.assign(
        aerosol_optical_thickness_spectral_355=ds[
            "aerosol_optical_thickness_spectral"
        ].isel({"aerosol_optical_thickness_dimension": 0}),
        aerosol_optical_thickness_spectral_670=ds[
            "aerosol_optical_thickness_spectral"
        ].isel({"aerosol_optical_thickness_dimension": 1}),
        aerosol_optical_thickness_spectral_865=ds[
            "aerosol_optical_thickness_spectral"
        ].isel({"aerosol_optical_thickness_dimension": 2}),
        aerosol_type_quality_1=ds["aerosol_type_quality"].isel(
            {"quality_indicator_dimension": 0}
        ),
        aerosol_type_quality_2=ds["aerosol_type_quality"].isel(
            {"quality_indicator_dimension": 1}
        ),
        aerosol_type_quality_3=ds["aerosol_type_quality"].isel(
            {"quality_indicator_dimension": 2}
        ),
        aerosol_angstrom_exponent_355_670=ds["aerosol_angstrom_exponent"].isel(
            {"angstrom_dimension": 0}
        ),
        aerosol_angstrom_exponent_670_865=ds["aerosol_angstrom_exponent"].isel(
            {"angstrom_dimension": 1}
        ),
    )

    ds = rename_var_info(
        ds,
        "aerosol_optical_thickness_spectral_355",
        long_name="AOT at 355 nm (AM-ACD)",
        units="",
    )
    ds = rename_var_info(
        ds,
        "aerosol_optical_thickness_spectral_670",
        long_name="AOT at 670 nm (M-AOT)",
        units="",
    )
    ds = rename_var_info(
        ds,
        "aerosol_optical_thickness_spectral_865",
        long_name="AOT at 865 nm (M-AOT)",
        units="",
    )
    ds = rename_var_info(
        ds,
        "quality_status",
        long_name="Quality status",
        units="",
    )
    ds = rename_var_info(
        ds,
        "aerosol_angstrom_exponent_355_670",
        long_name="Ångström exponent (355/670)",
        units="",
    )
    ds = rename_var_info(
        ds,
        "aerosol_angstrom_exponent_355_670",
        long_name="Ångström exponent (355/670)",
        units="",
    )

    return ds



import json
import urllib.parse as urlp
from logging import Logger

import requests
from lxml import html
from requests.cookies import RequestsCookieJar

from ._constants import URL_AUTHENTIFICATION
from ._exceptions import BadResponseError
from ._request import request_get, request_post


def _get_oads_login_url(server: str) -> str:
    return f"https://{server}/oads/access/login"


def get_oads_authentification_cookies(
    dissemination_server: str,
    username: str,
    password: str,
    logger: Logger | None = None,
) -> RequestsCookieJar:
    proxies: dict = {}

    # Requesting access to the dissemination server
    url_login = _get_oads_login_url(dissemination_server)
    response_login = request_get(url=url_login, logger=logger, proxies=proxies)
    cookies_login = response_login.cookies
    for r in response_login.history:
        cookies_login = requests.cookies.merge_cookies(cookies_login, r.cookies)

    tree_login = html.fromstring(response_login.content)
    _path = ".//input[@name = 'sessionDataKey']"
    session_data_key = tree_login.findall(_path)[0].attrib["value"]

    post_data = {
        "tocommonauth": "true",
        "username": username,
        "password": password,
        "sessionDataKey": session_data_key,
    }

    # Sending login request to the authentication platform
    url_auth = URL_AUTHENTIFICATION
    response_auth = request_post(
        url=url_auth,
        data=post_data,
        cookies=cookies_login,
        proxies=proxies,
    )
    tree_auth = html.fromstring(response_auth.content)

    # Extracting the variables needed to redirect from a successful authentication
    try:
        _path_relay_state = ".//input[@name='RelayState']"
        relay_state = tree_auth.findall(_path_relay_state)[0].attrib["value"]

        _path_saml_response = ".//input[@name='SAMLResponse']"
        saml_response = tree_auth.findall(_path_saml_response)[0].attrib["value"]
    except IndexError as e:
        exception_msg = "OADS did not responde as expected. Check your configuration file for valid a username and password."
        if logger:
            logger.exception(exception_msg)
        raise BadResponseError(exception_msg)

    # Defining the SAML redirection request
    post_data_saml = {
        "RelayState": relay_state,
        "SAMLResponse": saml_response,
    }
    url_saml = tree_auth.findall(".//form[@method='post']")[0].attrib["action"]
    response_saml = request_post(
        url=url_saml,
        data=post_data_saml,
        proxies=proxies,
        logger=logger,
    )
    cookies_saml = response_saml.cookies
    for r in response_saml.history:
        cookies_saml = requests.cookies.merge_cookies(cookies_saml, r.cookies)

    return cookies_saml



from logging import Logger

import requests

from ..utils._cli import console_exclusive_info, get_counter_message, log_textbox
from ..utils.config import ECKConfig
from ._auth_oads import get_oads_authentification_cookies
from ._eo_product import EOProduct, _DownloadResult
from ._eo_search_request import EOSearchRequest
from ._types import Entrypoint


def run_downloads(
    products: list[EOProduct],
    config: ECKConfig,
    entrypoint: Entrypoint,
    is_download: bool,
    is_overwrite: bool,
    is_unzip: bool,
    is_delete: bool,
    is_create_subdirs: bool,
    log_heading_msg: str = f"Download products",
    logger: Logger | None = None,
    is_reversed_order: bool = False,
) -> list[_DownloadResult]:
    if logger:
        console_exclusive_info()
        log_textbox(log_heading_msg, logger=logger, show_time=True)
        console_exclusive_info()

    if not is_download:
        if logger:
            logger.info(f"Skipped since option --no_download was used")
        return []

    if is_reversed_order:
        products.reverse()

    _current_server: str = ""
    _num_products: int = len(products)
    _download_results: list[_DownloadResult] = []
    oads_cookies_saml: requests.cookies.RequestsCookieJar | None = None
    for i, p in enumerate(products):
        if is_reversed_order:
            counter = _num_products - i
        else:
            counter = i + 1

        count_msg, _ = get_counter_message(counter, _num_products)

        if logger:
            if logger:
                logger.info(f"*{count_msg} Starting: {p.name}")
        if entrypoint == Entrypoint.OADS:
            if p.server != _current_server:
                if logger:
                    logger.info(
                        f" {count_msg} Authenticate at dissemination service: {p.server}"
                    )
                if len(config.oads_username) == 0 or len(config.oads_password) == 0:
                    msg = f"Authentication failed due to missing oads username or password"
                    raise ValueError(msg)
                oads_cookies_saml = get_oads_authentification_cookies(
                    dissemination_server=p.server,
                    username=config.oads_username,
                    password=config.oads_password,
                )
                _current_server = p.server
            _dlr = p.download(
                download_directory=config.path_to_data,
                is_overwrite=is_overwrite,
                is_unzip=is_unzip,
                is_delete=is_delete,
                is_create_subdirs=is_create_subdirs,
                oads_cookies_saml=oads_cookies_saml,
                total_count=_num_products,
                counter=counter,
                config=config,
                logger=logger,
            )
            _download_results.append(_dlr)
        elif entrypoint == Entrypoint.MAAP:
            _dlr = p.download(
                download_directory=config.path_to_data,
                is_overwrite=is_overwrite,
                is_unzip=is_unzip,
                is_delete=is_delete,
                is_create_subdirs=is_create_subdirs,
                maap_token=config.maap_token,
                total_count=_num_products,
                counter=counter,
                config=config,
                logger=logger,
            )
            _download_results.append(_dlr)
        else:
            raise ValueError(f"invalid value for entrypoint: {entrypoint}")
    return _download_results
